/*
 * G-code meta compiler
 *
 * Copyright (C) 2013  B. Stultiens
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "config.h"

#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <string.h>

#include "utils.h"
#include "grammartypes.h"
#include "value.h"
#include "builtin.h"
#include "builtin_gcode.h"
#include "builtin_svg.h"
#include "checkarg.h"

static const char *unitprintname(unit_et u)
{
	switch(u) {
	case UNIT_NONE:	return "none";
	case UNIT_MM:	return "mm";
	case UNIT_IN:	return "in";
	case UNIT_DEG:	return "degrees";
	case UNIT_RAD:	return "radians";
	}
	return "<unknown units>";
}

static value_t *bi_pi(const node_t *ref, int argc)
{
	check_arg_none(ref, "pi", argc);
	value_t *v = value_new(VAL_FLOAT);
	v->d = M_PI;
	return v;
}

static value_t *bi_position(const node_t *ref, int argc)
{
	int i;
	int nax = 3;	/* Default to XYZ */
	if(argc > 1)
		rtwarning(ref, "postion() has only one optional argument");
	if(argc > 0) {
		const value_t *v = value_get(argc, 1);
		if(v->unit != UNIT_NONE)
			rtwarning(ref, "postion() argument has units attached, ignored");
		nax = value_to_int(v);
		if(nax < 1)
			rterror(ref, "postion() argument must be > 0");
		if(nax > naxes)
			rtwarning(ref, "postion() argument returns maximum %d vector coordinates", naxes);
	}
	check_arg_none(ref, "position", argc);
	value_t *v = value_new(VAL_VECTOR);
	/* XXX: check if using rotational axes */
	for(i = 0; i < nax && i < naxes; i++)
		value_vector_add(v, value_new_flt(*global_pos[i], UNIT_MM));
	return v;
}

static value_t *bi_atan(const node_t *ref, int argc)
{
	double d0, d1;
	const value_t *arg0, *arg1;
	value_t *v;
	check_arg_two(ref, "atan", argc);
	arg0 = value_get(argc, 0);
	arg1 = value_get(argc, 1);
	check_arg_scalar(ref, "atan", arg0, 0);
	check_arg_scalar(ref, "atan", arg1, 1);
	check_arg_unit_distance(ref, "atan", arg0, 0);
	check_arg_unit_distance(ref, "atan", arg1, 1);
	d0 = value_to_mm(arg0);		/* Units don't matter as long as they are same for x and y in atan2(x, y) */
	d1 = value_to_mm(arg1);
	v = value_new(VAL_FLOAT);
	v->unit = UNIT_RAD;
	v->d = atan2(d0, d1);
	return v;
}

static value_t *bi_pow(const node_t *ref, int argc)
{
	double d0, d1;
	const value_t *arg0, *arg1;
	value_t *v;
	check_arg_two(ref, "pow", argc);
	arg0 = value_get(argc, 0);
	arg1 = value_get(argc, 1);
	check_arg_scalar(ref, "pow", arg0, 0);
	check_arg_scalar(ref, "pow", arg1, 1);
	check_arg_unit_none(ref, "pow", arg1, 1);
	d0 = value_to_double(arg0);
	d1 = value_to_double(arg1);
	v = value_new(VAL_FLOAT);
	v->d = pow(d0, d1);
	return v;
}

typedef enum {
	FN_sin,
	FN_cos,
	FN_tan,
	FN_asin,
	FN_acos,
	FN_round,
	FN_floor,
	FN_ceil,
	FN_loge,
	FN_log10,
	FN_log2,
	FN_exp,
	FN_sqrt,
} oneargfns_et;

static value_t *oneargdoubles(const node_t *ref, int argc, const char *fn, oneargfns_et idx)
{
	double d;
	const value_t *arg;
	value_t *v;
	check_arg_one(ref, fn, argc);
	arg = value_get(argc, 0);
	check_arg_scalar(ref, fn, arg, 0);
	switch(idx) {
	case FN_sin:
	case FN_cos:
	case FN_tan:
		check_arg_unit_angle(ref, fn, arg, 0);
		break;
	case FN_asin:
	case FN_acos:
		check_arg_unit_none(ref, fn, arg, 0);
		break;
	default:
		break;
	}
	d = value_to_double(arg);
	v = value_new(VAL_FLOAT);
	switch(idx) {
	case FN_sin:	v->d = sin(value_to_rad(arg)); break;
	case FN_cos:	v->d = cos(value_to_rad(arg)); break;
	case FN_tan:	v->d = tan(value_to_rad(arg)); break;
	case FN_asin:	check_range_double(ref, fn, 0, d, -1.0, 1.0); v->d = asin(d); v->unit = UNIT_RAD; break;
	case FN_acos:	check_range_double(ref, fn, 0, d, -1.0, 1.0); v->d = acos(d); v->unit = UNIT_RAD; break;
	case FN_round:	v->d = round(d); v->unit = arg->unit; break;
	case FN_floor:	v->d = floor(d); v->unit = arg->unit; break;
	case FN_ceil:	v->d = ceil(d); v->unit = arg->unit; break;
	case FN_loge:	check_range_double(ref, fn, 0, d, EPSILON, HUGE_VAL); v->d = log(d); break;
	case FN_log10:	check_range_double(ref, fn, 0, d, EPSILON, HUGE_VAL); v->d = log10(d); break;
	case FN_log2:	check_range_double(ref, fn, 0, d, EPSILON, HUGE_VAL); v->d = log2(d); break;
	case FN_exp:	v->d = exp(d); break;
	case FN_sqrt:	check_range_double(ref, fn, 0, d, 0.0, HUGE_VAL); v->d = sqrt(d); break;
	}
	return v;
}

#define ONEARGFNS(f)	static value_t *bi_ ## f (const node_t *ref, int argc) { return oneargdoubles(ref, argc, #f, FN_ ## f ); }

ONEARGFNS(sin)
ONEARGFNS(cos)
ONEARGFNS(tan)
ONEARGFNS(asin)
ONEARGFNS(acos)
ONEARGFNS(round)
ONEARGFNS(floor)
ONEARGFNS(ceil)
ONEARGFNS(loge)
ONEARGFNS(log10)
ONEARGFNS(log2)
ONEARGFNS(exp)
ONEARGFNS(sqrt)

static value_t *bi_int(const node_t *ref, int argc)
{
	int i, j;
	const value_t *arg;
	value_t *v = value_new(VAL_INT);
	check_arg_one(ref, "int", argc);
	arg = value_get(argc, 0);
	switch(arg->type) {
	case VAL_UNDEF:
		v->type = VAL_UNDEF;
		break;
	case VAL_FLOAT:
		v->i = arg->d;
		v->type = VAL_INT;
		break;
	case VAL_INT:
		v->i = arg->i;
		break;
	case VAL_VECTOR:
		v->type = VAL_VECTOR;
		for(i = 0; i < arg->v.n; i++) {
			switch(arg->v.vals[i]->type) {
			case VAL_UNDEF:	value_vector_add_undef(v); break;
			case VAL_FLOAT:	value_vector_add(v, value_new_int((int)arg->v.vals[i]->d, arg->v.vals[i]->unit)); break;
			case VAL_INT:	value_vector_add(v, value_new_int(arg->v.vals[i]->i, arg->v.vals[i]->unit)); break;
			default:
				rtinternal(ref, "int(): Invalid value type '%s' in vector[%d]", valuetypename(arg->v.vals[i]->type), i);
				break;
			}
		}
		break;
	case VAL_VECTORLIST:
		v->type = VAL_VECTORLIST;
		for(i = 0; i < arg->vl.n; i++) {
			const value_t *vec = arg->vl.vecs[i];
			value_t *nv = value_new(VAL_VECTOR);
			assert(vec->type == VAL_VECTOR);
			for(j = 0; j < vec->v.n; j++) {
				switch(vec->v.vals[j]->type) {
				case VAL_UNDEF:	value_vector_add_undef(nv); break;
				case VAL_FLOAT:	value_vector_add(nv, value_new_int((int)vec->v.vals[j]->d, vec->v.vals[j]->unit)); break;
				case VAL_INT:	value_vector_add(nv, value_new_int(vec->v.vals[j]->i, vec->v.vals[j]->unit)); break;
				default:
					rtinternal(ref, "int(): Invalid value type '%s' in vectorlist[%d][%d]", valuetypename(arg->v.vals[i]->type), i, j);
					break;
				}
			}
			value_vectorlist_add(v, nv);
		}
		break;
	case VAL_STRING:
	default:
		rterror(ref, "int(): Invalid argument value type '%s'", valuetypename(arg->type));
		break;
	}
	return v;
}

static value_t *bi_float(const node_t *ref, int argc)
{
	int i, j;
	const value_t *arg;
	value_t *v = value_new(VAL_FLOAT);
	check_arg_one(ref, "float", argc);
	arg = value_get(argc, 0);
	switch(arg->type) {
	case VAL_UNDEF:
		v->type = VAL_UNDEF;
		break;
	case VAL_FLOAT:
		v->d = arg->d;
		break;
	case VAL_INT:
		v->d = arg->i;
		v->type = VAL_FLOAT;
		break;
	case VAL_VECTOR:
		v->type = VAL_VECTOR;
		for(i = 0; i < arg->v.n; i++) {
			switch(arg->v.vals[i]->type) {
			case VAL_UNDEF:	value_vector_add_undef(v); break;
			case VAL_FLOAT:	value_vector_add(v, value_new_flt(arg->v.vals[i]->d, arg->v.vals[i]->unit)); break;
			case VAL_INT:	value_vector_add(v, value_new_flt((double)arg->v.vals[i]->i, arg->v.vals[i]->unit)); break;
			default:
				rtinternal(ref, "float(): Invalid value type '%s' in vector[%d]", valuetypename(arg->v.vals[i]->type), i);
				break;
			}
		}
		break;
	case VAL_VECTORLIST:
		v->type = VAL_VECTORLIST;
		for(i = 0; i < arg->vl.n; i++) {
			const value_t *vec = arg->vl.vecs[i];
			value_t *nv = value_new(VAL_VECTOR);
			assert(vec->type == VAL_VECTOR);
			for(j = 0; j < vec->v.n; j++) {
				switch(vec->v.vals[j]->type) {
				case VAL_UNDEF:	value_vector_add_undef(nv); break;
				case VAL_FLOAT:	value_vector_add(nv, value_new_flt(vec->v.vals[j]->d, vec->v.vals[j]->unit)); break;
				case VAL_INT:	value_vector_add(nv, value_new_flt((double)vec->v.vals[j]->i, vec->v.vals[j]->unit)); break;
				default:
					rtinternal(ref, "float(): Invalid value type '%s' in vectorlist[%d][%d]", valuetypename(arg->v.vals[i]->type), i, j);
					break;
				}
			}
			value_vectorlist_add(v, nv);
		}
		break;
	case VAL_STRING:
	default:
		rterror(ref, "float(): Invalid argument value type '%s'", valuetypename(arg->type));
		break;
	}
	return v;
}

static value_t *bi_abs(const node_t *ref, int argc)
{
	const value_t *arg;
	value_t *v;
	check_arg_one(ref, "abs", argc);
	arg = value_get(argc, 0);
	check_arg_scalar(ref, "abs", arg, 0);
	v = value_new(arg->type);
	switch(arg->type) {
	case VAL_INT:	v->i = abs(arg->i); break;
	case VAL_FLOAT:	v->d = fabs(arg->d); break;
	default:	rterror(ref, "Argument to abs() should be scalar"); break;
	}
	v->unit = arg->unit;
	return v;
}

typedef value_t *(*cvtfunc_t)(const node_t *ref, const value_t *arg);

static value_t *_to_mm(const node_t *ref, const value_t *arg)
{
	value_t *v = value_new(arg->type);
	if(arg->type == VAL_UNDEF)
		return v;
	assert(arg->type == VAL_INT || arg->type == VAL_FLOAT);
	switch(arg->unit) {
	case UNIT_DEG:
	case UNIT_RAD:
		rtwarning(ref, "Conversion from %s to mm not compatible", unitprintname(arg->unit));
		/* Fallthrough */
	case UNIT_NONE:
	case UNIT_MM:
		if(v->type == VAL_INT)
			v->i = arg->i;
		else
			v->d = arg->d;
		break;
	case UNIT_IN:
		if(v->type == VAL_INT)
			v->d = (double)arg->i * 25.4;
		else
			v->d = arg->d * 25.4;
		v->type = VAL_FLOAT;
		break;
	}
	v->unit = UNIT_MM;
	return v;
}

static value_t *_to_in(const node_t *ref, const value_t *arg)
{
	value_t *v = value_new(arg->type);
	if(arg->type == VAL_UNDEF)
		return v;
	assert(arg->type == VAL_INT || arg->type == VAL_FLOAT);
	switch(arg->unit) {
	case UNIT_DEG:
	case UNIT_RAD:
		rtwarning(ref, "Conversion from %s to inch not compatible", unitprintname(arg->unit));
		/* Fallthrough */
	case UNIT_NONE:
	case UNIT_IN:
		if(v->type == VAL_INT)
			v->i = arg->i;
		else
			v->d = arg->d;
		break;
	case UNIT_MM:
		if(v->type == VAL_INT)
			v->d = (double)arg->i / 25.4;
		else
			v->d = arg->d / 25.4;
		v->type = VAL_FLOAT;
		break;
	}
	v->unit = UNIT_IN;
	return v;
}

static value_t *_to_deg(const node_t *ref, const value_t *arg)
{
	value_t *v = value_new(arg->type);
	if(arg->type == VAL_UNDEF)
		return v;
	assert(arg->type == VAL_INT || arg->type == VAL_FLOAT);
	switch(arg->unit) {
	case UNIT_MM:
	case UNIT_IN:
		rtwarning(ref, "Conversion from %s to degrees not compatible", unitprintname(arg->unit));
		/* Fallthrough */
	case UNIT_NONE:
	case UNIT_DEG:
		if(v->type == VAL_INT)
			v->i = arg->i;
		else
			v->d = arg->d;
		break;
	case UNIT_RAD:
		if(v->type == VAL_INT)
			v->d = (double)arg->i / M_PI * 180.0;
		else
			v->d = arg->d / M_PI * 180.0;
		v->type = VAL_FLOAT;
		break;
	}
	v->unit = UNIT_DEG;
	return v;
}

static value_t *_to_rad(const node_t *ref, const value_t *arg)
{
	value_t *v = value_new(arg->type);
	if(arg->type == VAL_UNDEF)
		return v;
	assert(arg->type == VAL_INT || arg->type == VAL_FLOAT);
	switch(arg->unit) {
	case UNIT_MM:
	case UNIT_IN:
		rtwarning(ref, "Conversion from %s to degrees not compatible", unitprintname(arg->unit));
		/* Fallthrough */
	case UNIT_NONE:
	case UNIT_RAD:
		if(v->type == VAL_INT)
			v->i = arg->i;
		else
			v->d = arg->d;
		break;
	case UNIT_DEG:
		if(v->type == VAL_INT)
			v->d = (double)arg->i * M_PI / 180.0;
		else
			v->d = arg->d * M_PI / 180.0;
		v->type = VAL_FLOAT;
		break;
	}
	v->unit = UNIT_RAD;
	return v;
}

static value_t *_to_none(const node_t *ref, const value_t *arg)
{
	value_t *v = value_new(arg->type);
	(void)ref;
	assert(v->type == VAL_INT || v->type == VAL_FLOAT);
	if(v->type == VAL_INT)
		v->i = arg->i;
	else
		v->d = arg->d;
	v->unit = UNIT_NONE;
	return v;
}

static value_t *unit_conversion(const node_t *ref, const char *fn, int argc, cvtfunc_t cvt)
{
	const value_t *arg;
	value_t *v;
	int i, j;
	check_arg_one(ref, fn, argc);
	arg = value_get(argc, 0);
	switch(arg->type) {
	case VAL_INT:
	case VAL_FLOAT:
		return cvt(ref, arg);
	case VAL_VECTOR:
		v = value_new(VAL_VECTOR);
		for(i = 0; i < arg->v.n; i++)
			value_vector_add(v, cvt(ref, arg->v.vals[i]));
		return v;
	case VAL_VECTORLIST:
		v = value_new(VAL_VECTORLIST);
		for(i = 0; i < arg->vl.n; i++) {
			value_t *vec = arg->vl.vecs[i];
			value_t *newvec = value_new(VAL_VECTOR);
			for(j = 0; j < vec->v.n; j++)
				value_vector_add(newvec, cvt(ref, vec->v.vals[j]));
			value_vectorlist_add(v, newvec);
		}
		return v;
	case VAL_UNDEF:
		break;
	case VAL_STRING:
		rterror(ref, "Strings cannot be converted to value with units in %s()", fn);
		break;
	default:
		rtinternal(ref, "Undefined value type '%d' in %s()", arg->type, fn);
		break;
	}
	return value_new(VAL_UNDEF);
}

static value_t *bi_to_mm(const node_t *ref, int argc)
{
	return unit_conversion(ref, "to_mm", argc, _to_mm);
}

static value_t *bi_to_in(const node_t *ref, int argc)
{
	return unit_conversion(ref, "to_in", argc, _to_in);
}

static value_t *bi_to_deg(const node_t *ref, int argc)
{
	return unit_conversion(ref, "to_deg", argc, _to_deg);
}

static value_t *bi_to_rad(const node_t *ref, int argc)
{
	return unit_conversion(ref, "to_rad", argc, _to_rad);
}

static value_t *bi_to_none(const node_t *ref, int argc)
{
	return unit_conversion(ref, "to_none", argc, _to_none);
}

static value_t *bi_count(const node_t *ref, int argc)
{
	const value_t *arg;
	value_t *v;
	check_arg_one(ref, "count", argc);
	arg = value_get(argc, 0);
	v = value_new(VAL_INT);
	v->unit = UNIT_NONE;
	switch(arg->type) {
	case VAL_VECTOR:	v->i = arg->v.n; break;
	case VAL_VECTORLIST:	v->i = arg->vl.n; break;
	case VAL_UNDEF:
	case VAL_INT:
	case VAL_FLOAT:
	case VAL_STRING:
		rterror(ref, "count() can only count vectors and vectorlists");
		break;
	}
	return v;
}

static void printunit(FILE *fp, unit_et u)
{
	switch(u) {
	case UNIT_NONE:	break;
	case UNIT_MM:	fprintf(fp, "mm"); break;
	case UNIT_IN:	fprintf(fp, "in"); break;
	case UNIT_DEG:	fprintf(fp, "deg"); break;
	case UNIT_RAD:	fprintf(fp, "rad"); break;
	}
}

static void printvector(FILE *fp, const node_t *ref, const value_t *v)
{
	int i;
	assert(v->type == VAL_VECTOR);
	fprintf(fp, "[");
	for(i = 0; i < v->v.n; i++) {
		if(i)
			fprintf(fp, ",");
		switch(v->v.vals[i]->type) {
		case VAL_UNDEF:		fprintf(fp, "-"); break;
		case VAL_INT:		fprintf(fp, "%d", v->v.vals[i]->i); printunit(fp, v->v.vals[i]->unit); break;
		case VAL_FLOAT:		fprintf(fp, "%.*f", cl_decimals, v->v.vals[i]->d); printunit(fp, v->v.vals[i]->unit); break;
		default:		rtinternal(ref, "Invalid value type %d in vector printing", v->v.vals[i]->type);
		}
	}
	fprintf(fp, "]");
}

static value_t *printfunc(const node_t *ref, int argc, const char *pfx, FILE *fp, int forcenl)
{
	int i, j;
	int seennl = 0;

	if(pfx)
		fprintf(fp, "%s:%d:%d: %s: ", ref->filename, ref->linenr, ref->charnr, pfx);

	for(i = 0; i < argc; i++) {
		const value_t *v = value_get(argc, i);
		if(i == argc-1 && v->type == VAL_STRING && v->str[0] && v->str[strlen(v->str)-1] == '\n')
			seennl = 1;
		switch(v->type) {
		case VAL_UNDEF:		fprintf(fp, "<undef>"); break;
		case VAL_INT:		fprintf(fp, "%d", v->i); printunit(fp, v->unit); break;
		case VAL_FLOAT:		fprintf(fp, "%.*f", cl_decimals, v->d); printunit(fp, v->unit); break;
		case VAL_STRING:	fprintf(fp, "%s", v->str); break;
		case VAL_VECTOR:	printvector(fp, ref, v); break;
		case VAL_VECTORLIST:
			fprintf(fp, "{");
			for(j = 0; j < v->vl.n; j++) {
				if(j)
					fprintf(fp, ",");
				printvector(fp, ref, v->vl.vecs[j]);
			}
			fprintf(fp, "}");
			break;
		}
	}

	if(forcenl && !seennl)
		fprintf(fp, "\n");
	return value_new(VAL_UNDEF);
}

static value_t *bi_literal(const node_t *ref, int argc)
{
	return printfunc(ref, argc, NULL, ofp, 0);
}

static value_t *bi_error(const node_t *ref, int argc)
{
	runtimeerrors++;
	return printfunc(ref, argc, "Runtime error()", stderr, 1);
}

static value_t *bi_warning(const node_t *ref, int argc)
{
	runtimewarnings++;
	return printfunc(ref, argc, "Runtime warning()", stderr, 1);
}

static value_t *bi_message(const node_t *ref, int argc)
{
	return printfunc(ref, argc, "Runtime message()", stderr, 1);
}

static value_t *bi_normalize(const node_t *ref, int argc)
{
	int i;
	const value_t *arg;
	value_t *v;
	double sum = 0.0;
	check_arg_one(ref, "normalize", argc);
	arg = value_get(argc, 0);
	check_arg_vector(ref, "normalize", arg, 0);

	/* Units in normalize() are not important as long as they are the same for all */
	for(i = 0; i < arg->v.n; i++) {
		double d = value_to_mm(arg->v.vals[i]);
		check_arg_unit_distance(ref, "normalize", arg->v.vals[i], 0);
		sum += d*d;
	}
	sum = sqrt(sum);
	v = value_new(VAL_VECTOR);
	for(i = 0; i < arg->v.n; i++) {
		switch(arg->v.vals[i]->type) {
		case VAL_UNDEF:
			value_vector_add_undef(v);
			break;
		case VAL_INT:
			value_vector_add(v, value_new_flt(value_to_mm(arg->v.vals[i]) / sum, UNIT_NONE));
			break;
		case VAL_FLOAT:
			value_vector_add(v, value_new_flt(value_to_mm(arg->v.vals[i]) / sum, UNIT_NONE));
			break;
		default:
			rtinternal(ref, "Vector value contains non-scalar elements");
			break;
		}
		
	}
	return v;
}

static value_t *bi_length(const node_t *ref, int argc)
{
	int i;
	const value_t *arg;
	double sum = 0.0;
	unit_et u = UNIT_NONE;
	check_arg_one(ref, "length", argc);
	arg = value_get(argc, 0);
	check_arg_vector(ref, "length", arg, 0);

	for(i = 0; i < arg->v.n; i++) {
		if(arg->v.vals[i]->type == VAL_UNDEF)
			continue;
		assert(isscalar(arg->v.vals[i]));
		if(u == UNIT_NONE) {
			if(isdistance(arg->v.vals[i]))
				u = cl_inch ? UNIT_IN : UNIT_MM;
			else if(isangle(arg->v.vals[i]))
				u = UNIT_DEG;
		} else if(((u == UNIT_MM || u == UNIT_IN) && !isdistance(arg->v.vals[i])) || (u == UNIT_DEG && !isangle(arg->v.vals[i]))) {
			rtwarning(ref, "Mix of distance and angular units in vector in length() operation");
			break;
		}
	}
	for(i = 0; i < arg->v.n; i++) {
		double d;
		switch(u) {
		case UNIT_MM:
			d = value_to_mm(arg->v.vals[i]);
			break;
		case UNIT_IN:
			d = value_to_in(arg->v.vals[i]);
			break;
		case UNIT_DEG:
			d = value_to_deg(arg->v.vals[i]);
			break;
		case UNIT_NONE:
		default:
			d = value_to_double(arg->v.vals[i]);
			break;
		}
		sum += d*d;
	}
	return value_new_flt(sqrt(sum), u);
}

typedef enum {
	FN_rotate_xy,
	FN_rotate_xz,
	FN_rotate_yz,
} rot_et;

static inline double getcoord(const value_t *vec, int coord, unit_et u)
{
	if(coord >= vec->v.n)
		return 0.0;
	switch(u) {
	case UNIT_MM:	return value_to_mm(vec->v.vals[coord]);
	case UNIT_IN:	return value_to_in(vec->v.vals[coord]);
	default:	return value_to_double(vec->v.vals[coord]);
	}
}

static inline unit_et getunit(const node_t *ref, const value_t *vec, int coord, const char *fn)
{
	const char ax[3] = "XYZ";
	assert(coord < 3);
	if(coord >= vec->v.n || vec->v.vals[coord]->type == VAL_UNDEF)
		return UNIT_NONE;
	if(!isdistance(vec->v.vals[coord])) {
		rtwarning(ref, "%s() Vector coordinate %d (%c) is not a distance", fn, coord, ax[coord]);
		return UNIT_NONE;
	}
	return vec->v.vals[coord]->unit;
}

static value_t *rotate_vec(const node_t  *ref, const value_t *vec, const value_t *ang, const char *fn, rot_et ft)
{
	double x, y, phi;
	double xr, yr;
	unit_et ux, uy, ut;
	value_t *v = value_dup(vec);
	assert(isvector(vec));
	assert(isscalar(ang));

	phi = value_to_rad(ang);

	switch(ft) {
	case FN_rotate_xy:
		while(v->v.n < 2)
			value_vector_add_undef(v);
		if(vec->v.n < 1)
			rtwarning(ref, "%s() Vector missing all coordinates to rotate", fn);
		ux = getunit(ref, vec, 0, fn);
		uy = getunit(ref, vec, 1, fn);
		if(ux == uy || uy == UNIT_NONE)
			ut = ux;
		else
			ut = uy;
		x = getcoord(vec, 0, ut);
		y = getcoord(vec, 1, ut);
		xr = x * cos(phi) - y * sin(phi);
		yr = x * sin(phi) + y * cos(phi);
		v->v.vals[0]->type = VAL_FLOAT;
		v->v.vals[0]->unit = ut;
		v->v.vals[0]->d = xr;
		v->v.vals[1]->type = VAL_FLOAT;
		v->v.vals[1]->unit = ut;
		v->v.vals[1]->d = yr;
		break;
	case FN_rotate_xz:
		while(v->v.n < 3)
			value_vector_add_undef(v);
		if(vec->v.n < 1)
			rtwarning(ref, "%s() Vector missing all coordinates to rotate", fn);
		ux = getunit(ref, vec, 0, fn);
		uy = getunit(ref, vec, 2, fn);
		if(ux == uy || uy == UNIT_NONE)
			ut = ux;
		else
			ut = uy;
		x = getcoord(vec, 0, ut);
		y = getcoord(vec, 2, ut);
		xr = x * cos(phi) - y * sin(phi);
		yr = x * sin(phi) + y * cos(phi);
		v->v.vals[0]->type = VAL_FLOAT;
		v->v.vals[0]->unit = ut;
		v->v.vals[0]->d = xr;
		v->v.vals[2]->type = VAL_FLOAT;
		v->v.vals[2]->unit = ut;
		v->v.vals[2]->d = yr;
		break;
	case FN_rotate_yz:
		while(v->v.n < 3)
			value_vector_add_undef(v);
		if(vec->v.n < 2)
			rtwarning(ref, "%s() Vector missing all coordinates to rotate", fn);
		ux = getunit(ref, vec, 1, fn);
		uy = getunit(ref, vec, 2, fn);
		if(ux == uy || uy == UNIT_NONE)
			ut = ux;
		else
			ut = uy;
		x = getcoord(vec, 1, ut);
		y = getcoord(vec, 2, ut);
		xr = x * cos(phi) - y * sin(phi);
		yr = x * sin(phi) + y * cos(phi);
		v->v.vals[1]->type = VAL_FLOAT;
		v->v.vals[1]->unit = ut;
		v->v.vals[1]->d = xr;
		v->v.vals[2]->type = VAL_FLOAT;
		v->v.vals[2]->unit = ut;
		v->v.vals[2]->d = yr;
		break;
	default:
		rtinternal(ref, "Invalid target function reference %d in rotate_vec()", ft);
	}
	return v;
}

static value_t *rotateany(const node_t *ref, int argc, const char *fn, rot_et ft)
{
	const value_t *arg0;
	const value_t *arg1;
	check_arg_two(ref, fn, argc);
	arg0 = value_get(argc, 0);
	arg1 = value_get(argc, 1);
	if(arg0->type != VAL_VECTOR && arg0->type != VAL_VECTORLIST) {
		rterror(ref, "%s() argument 1 must be vector or vectorlist", fn);
		return value_new(VAL_UNDEF);
	}
	check_arg_scalar(ref, fn, arg1, 1);
	check_arg_unit_angle(ref, fn, arg1, 1);

	if(arg0->type == VAL_VECTOR)
		return rotate_vec(ref, arg0, arg1, fn, ft);
	else {
		int i;
		value_t *v = value_new(VAL_VECTORLIST);
		for(i = 0; i < arg0->vl.n; i++)
			value_vectorlist_add(v, rotate_vec(ref, arg0->vl.vecs[i], arg1, fn, ft));
		return v;
	}
}

static value_t *bi_rotate_xy(const node_t *ref, int argc)
{
	return rotateany(ref, argc, "rotate_xy", FN_rotate_xy);
}

static value_t *bi_rotate_xz(const node_t *ref, int argc)
{
	return rotateany(ref, argc, "rotate_xz", FN_rotate_xz);
}

static value_t *bi_rotate_yz(const node_t *ref, int argc)
{
	return rotateany(ref, argc, "rotate_yz", FN_rotate_yz);
}

typedef enum {
	FN_isfloat,
	FN_isint,
	FN_isscalar,
	FN_isstring,
	FN_isundef,
	FN_isvector,
	FN_isvectorlist,
} boolfunc_et;

#define BOOLFUNC(f)	static value_t *bi_ ## f(const node_t *ref, int argc) { return boolfuncs(ref, argc, #f, FN_ ## f); }

static value_t *boolfuncs(const node_t *ref, int argc, const char *fn, boolfunc_et ft)
{
	const value_t *arg;
	value_t *v = value_new(VAL_INT);
	check_arg_one(ref, fn, argc);
	arg = value_get(argc, 0);
	switch(ft) {
	case FN_isfloat:	v->i = arg->type == VAL_FLOAT; break;
	case FN_isint:		v->i = arg->type == VAL_INT; break;
	case FN_isscalar:	v->i = arg->type == VAL_FLOAT || arg->type == VAL_INT; break;
	case FN_isstring:	v->i = arg->type == VAL_STRING; break;
	case FN_isundef:	v->i = arg->type == VAL_UNDEF; break;
	case FN_isvector:	v->i = arg->type == VAL_VECTOR; break;
	case FN_isvectorlist:	v->i = arg->type == VAL_VECTORLIST; break;
	}
	return v;
}

BOOLFUNC(isfloat)
BOOLFUNC(isint)
BOOLFUNC(isscalar)
BOOLFUNC(isstring)
BOOLFUNC(isundef)
BOOLFUNC(isvector)
BOOLFUNC(isvectorlist)

typedef enum {
	FN_isnone,
	FN_isangle,
	FN_isdistance,
	FN_isdeg,
	FN_israd,
	FN_ismm,
	FN_isinch,
} unitfunc_et;

static value_t *unitfuncs(const node_t *ref, int argc, const char *fn, unitfunc_et ft)
{
	const value_t *arg;
	value_t *v = value_new(VAL_INT);
	check_arg_one(ref, fn, argc);
	arg = value_get(argc, 0);
	check_arg_scalar(ref, fn, arg, 0);
	if(isscalar(arg)) {
		switch(ft) {
		case FN_isnone:		v->i = arg->unit == UNIT_NONE; break;
		case FN_isangle:	v->i = arg->unit == UNIT_NONE || arg->unit == UNIT_DEG || arg->unit == UNIT_RAD; break;
		case FN_isdistance:	v->i = arg->unit == UNIT_NONE || arg->unit == UNIT_MM || arg->unit == UNIT_IN; break;
		case FN_isdeg:		v->i = arg->unit == UNIT_DEG; break;
		case FN_israd:		v->i = arg->unit == UNIT_RAD; break;
		case FN_ismm:		v->i = arg->unit == UNIT_MM; break;
		case FN_isinch:		v->i = arg->unit == UNIT_IN; break;
		}
	} else
		v->i = 0;
	return v;
}

#define UNITFUNC(f)	static value_t *bi_ ## f(const node_t *ref, int argc) { return unitfuncs(ref, argc, #f, FN_ ## f); }

UNITFUNC(isnone)
UNITFUNC(isangle)
UNITFUNC(isdistance)
UNITFUNC(isdeg)
UNITFUNC(israd)
UNITFUNC(ismm)
UNITFUNC(isinch)


static value_t *bi_head(const node_t *ref, int argc)
{
	int i;
	const value_t *arg0;
	const value_t *arg1;
	value_t *v;
	int count;
	check_arg_two(ref, "head", argc);
	arg0 = value_get(argc, 0);
	if(!isvector(arg0) && !isvectorlist(arg0)) {
		rterror(ref, "head() argument 0 must be vector or vectorlist");
		return value_new(VAL_UNDEF);
	}
	arg1 = value_get(argc, 1);
	check_arg_scalar(ref, "head", arg1, 1);
	check_arg_unit_none(ref, "head", arg1, 1);
	count = value_to_int(arg1);

	if(count < 0) {
		rtwarning(ref, "head() count must be >=0, returning empty");
		count = 0;
	}

	v = value_new(arg0->type);

	/* Zero count --> empty */
	if(!count)
		return v;

	if(isvector(arg0)) {
		for(i = 0; i < count && i < arg0->v.n; i++)
			value_vector_add(v, value_dup(arg0->v.vals[i]));
		while(count < v->v.n)
			value_vector_add_undef(v);
	} else {
		for(i = 0; i < count && i < arg0->vl.n; i++)
			value_vectorlist_add(v, value_dup(arg0->vl.vecs[i]));
	}
	return v;
}

static value_t *bi_tail(const node_t *ref, int argc)
{
	int i;
	const value_t *arg0;
	const value_t *arg1;
	value_t *v;
	int count;
	check_arg_two(ref, "tail", argc);
	arg0 = value_get(argc, 0);
	if(!isvector(arg0) && !isvectorlist(arg0)) {
		rterror(ref, "tail() argument 0 must be vector or vectorlist");
		return value_new(VAL_UNDEF);
	}
	arg1 = value_get(argc, 1);
	check_arg_scalar(ref, "tail", arg1, 1);
	check_arg_unit_none(ref, "tail", arg1, 1);
	count = value_to_int(arg1);

	if(count < 0) {
		rtwarning(ref, "tail() count must be >=0, returning empty");
		count = 0;
	}

	v = value_new(arg0->type);

	/* Zero count --> empty */
	if(!count)
		return v;

	if(isvector(arg0)) {
		int n = arg0->v.n;
		while(n < count) {
			value_vector_add_undef(v);
			count--;
		}
		i = n > count ? n - count : 0;
		for(; count && i < n; i++, count--)
			value_vector_add(v, value_dup(arg0->v.vals[i]));
	} else {
		int n = arg0->v.n;
		i = n > count ? n - count : 0;
		for(; count && i < n; i++, count--)
			value_vectorlist_add(v, value_dup(arg0->vl.vecs[i]));
	}
	return v;
}

value_t *bi_plane(const node_t *ref, int argc)
{
	int i = -1;
	const value_t *arg;
	check_arg_one(ref, "plane", argc);
	arg = value_get(argc, 0);
	if(isscalar(arg)) {
		check_arg_unit_none(ref, "plane", arg, 0);
		i = value_to_int(arg);
		if(i < 0 || i > 2)
			goto arg_error;
	} else if(isstring(arg)) {
		if(!strcasecmp(arg->str, "xy"))
			i = PLANE_XY;
		else if(!strcasecmp(arg->str, "xz"))
			i = PLANE_XZ;
		else if(!strcasecmp(arg->str, "yz"))
			i = PLANE_YZ;
		else
			goto arg_error;
	} else {
arg_error:
		rterror(ref, "plane() only accepts 0..2 or string XY|XZ|YZ");
	}
	assert(i >= 0 && i <= 2);
	global_plane = i;
	return value_new(VAL_UNDEF);
}

static value_t *vecscalemul(const node_t *ref, const value_t *lv, const value_t *rv)
{
	int i;
	value_t *v = value_new(VAL_VECTOR);
	for(i = 0; i < lv->v.n || i < rv->v.n; i++) {
		if(i < lv->v.n) {
			if(i < rv->v.n && !isundef(rv->v.vals[i])) {
				value_vector_add(v, op_mul(ref, value_dup(lv->v.vals[i]), rv->v.vals[i]));
			} else
				value_vector_add(v, value_dup(lv->v.vals[i]));
		} else
			value_vector_add(v, value_new(VAL_UNDEF));
	}
	return v;
}

static value_t *bi_scale(const node_t *ref, int argc)
{
	int i;
	const value_t *arg0;
	const value_t *arg1;
	check_arg_two(ref, "scale", argc);
	arg0 = value_get(argc, 0);
	if(!isvector(arg0) && !isvectorlist(arg0)) {
		rterror(ref, "scale() argument 1 must be vector or vectorlist");
		return value_new(VAL_UNDEF);
	}
	arg1 = value_get(argc, 1);
	check_arg_vector(ref, "scale", arg1, 1);

	if(isvector(arg0)) {
		return vecscalemul(ref, arg0, arg1);
	} else if(isvectorlist(arg0)) {
		value_t *v = value_new(VAL_VECTORLIST);
		for(i = 0; i < arg0->vl.n; i++)
			value_vectorlist_add(v, vecscalemul(ref, arg0->vl.vecs[i], arg1));
		return v;
	}
	return value_new(VAL_UNDEF);
}

static value_t *bi_isgcode(const node_t *ref, int argc)
{
	value_t *v = value_new(VAL_INT);
	check_arg_none(ref, "isgcode", argc);
	v->i = cl_format == FMT_GCODE;
	return v;
}

static value_t *bi_issvg(const node_t *ref, int argc)
{
	value_t *v = value_new(VAL_INT);
	check_arg_none(ref, "issvg", argc);
	v->i = cl_format == FMT_SVG;
	return v;
}

static value_t *bi_isdxf(const node_t *ref, int argc)
{
	value_t *v = value_new(VAL_INT);
	check_arg_none(ref, "isdxf", argc);
	v->i = cl_format == FMT_DXF;
	return v;
}

static value_t *bi_dummy(const node_t *ref, int argc)
{
	(void)ref;
	(void)argc;
	return value_new(VAL_UNDEF);
}

const builtins_t builtins_gcmc[] = {
	{ "abs",		bi_abs },
	{ "acos",		bi_acos },
	{ "asin",		bi_asin },
	{ "atan",		bi_atan },
	{ "ceil",		bi_ceil },
	{ "cos",		bi_cos },
	{ "count",		bi_count },
	{ "error",		bi_error },
	{ "exp",		bi_exp },
	{ "float",		bi_float },
	{ "floor",		bi_floor },
	{ "head",		bi_head },
	{ "int",		bi_int },
	{ "isangle",		bi_isangle },
	{ "isdeg",		bi_isdeg },
	{ "isdistance",		bi_isdistance },
	{ "isdxf",		bi_isdxf },
	{ "isfloat",		bi_isfloat },
	{ "isgcode",		bi_isgcode },
	{ "isinch",		bi_isinch },
	{ "isint",		bi_isint },
	{ "ismm",		bi_ismm },
	{ "isnone",		bi_isnone },
	{ "israd",		bi_israd },
	{ "isscalar",		bi_isscalar },
	{ "isstring",		bi_isstring },
	{ "issvg",		bi_issvg },
	{ "isundef",		bi_isundef },
	{ "isvector",		bi_isvector },
	{ "isvectorlist",	bi_isvectorlist },
	{ "length",		bi_length },
	{ "literal",		bi_literal },
	{ "log10",		bi_log10 },
	{ "log2",		bi_log2 },
	{ "loge",		bi_loge },
	{ "message",		bi_message },
	{ "normalize",		bi_normalize },
	{ "pi",			bi_pi },
	{ "plane",		bi_plane },	/* Called from the specific backends */
	{ "position",		bi_position },
	{ "pow",		bi_pow },
	{ "rotate_xy",		bi_rotate_xy },
	{ "rotate_xz",		bi_rotate_xz },
	{ "rotate_yz",		bi_rotate_yz },
	{ "round",		bi_round },
	{ "scale",		bi_scale },
	{ "sin",		bi_sin },
	{ "sqrt",		bi_sqrt },
	{ "svg_closepath",	bi_dummy },	/* Override in builtin_cvg */
	{ "tail",		bi_tail },
	{ "tan",		bi_tan },
	{ "to_deg",		bi_to_deg },
	{ "to_in",		bi_to_in },
	{ "to_mm",		bi_to_mm },
	{ "to_none",		bi_to_none },
	{ "to_rad",		bi_to_rad },
	{ "warning",		bi_warning },
};

const int nbuiltins_gcmc = NELEM(builtins_gcmc);
